@@ -9,9 +9,12 @@ Key Features:
 - Memory-efficient batch collation
 - Cross-platform compatibility (Windows/Linux)
 - Performance monitoring and statistics
+- Load balancing across worker processes
+- CUDA-safe spawn context for GPU training
+- Reusable pinned memory pools for efficiency
 """
 
-from neurograd import Tensor, float32, xp
+from neurograd import float32, xp
 import math
 import random
 import os
@@ -19,7 +22,11 @@ import cv2
 cv2.setNumThreads(1)  # Prevent OpenCV threading conflicts
 import numpy as np
 from typing import Optional, Union, Callable, Any
-import multiprocessing as mp
+try:
+    import multiprocess as mp  # Use multiprocess (cloudpickle-enabled) instead of multiprocessing
+except ImportError:
+    import multiprocessing as mp  # Fallback to standard multiprocessing
+    print("Warning: multiprocess not available, falling back to multiprocessing. Some functions may not be picklable.")
 import queue
 import time
 import threading
@@ -37,6 +44,8 @@ except ImportError:
 class Dataset:
     def __init__(self, X, y, dtype=float32):
         assert len(X) == len(y), "Mismatched input and label lengths"
+        # Import Tensor locally to avoid serialization issues
+        from neurograd import Tensor
         self.X = Tensor(X, dtype=dtype)
         self.y = Tensor(y, dtype=dtype)
     
@@ -70,38 +79,46 @@ class Dataset:
 
 
 class DataLoader:
-
+    # Class-level pinned memory pool for reuse across instances
+    _pinned_pool = None
+    _pinned_cache = {}  # Cache pinned arrays by size/dtype
+    
     @staticmethod
     def _worker_loop(dataset, index_queue, data_queue, worker_id, pin_memory=False):
         """Worker process function for loading data samples."""
         # Set OpenCV threading for worker process
         cv2.setNumThreads(1)
         
-        while True:
-            try:
-                indices = index_queue.get(timeout=1)
-                if indices is None:  # Shutdown signal
+        try:
+            while True:
+                try:
+                    indices = index_queue.get(timeout=1)
+                    if indices is None:  # Shutdown signal
+                        break
+                        
+                    # Process batch
+                    batch_data = []
+                    for idx in indices:
+                        try:
+                            sample = dataset[idx]
+                            batch_data.append(sample)
+                        except Exception as e:
+                            print(f"Worker {worker_id}: Error loading sample {idx}: {e}")
+                            continue
+                    
+                    if batch_data:
+                        data_queue.put((worker_id, indices, batch_data))
+                        
+                except queue.Empty:
+                    continue
+                except (EOFError, BrokenPipeError):
                     break
-                    
-                # Load batch data
-                batch_data = []
-                for idx in indices:
-                    try:
-                        sample = dataset[idx]
-                        batch_data.append(sample)
-                    except Exception as e:
-                        # Handle corrupted data gracefully
-                        print(f"Worker {worker_id}: Error loading sample {idx}: {e}")
-                        continue
-                
-                if batch_data:
-                    data_queue.put((indices, batch_data))
-                    
-            except queue.Empty:
-                continue
-            except Exception as e:
-                print(f"Worker {worker_id}: Unexpected error: {e}")
-                break
+                except Exception as e:
+                    print(f"Worker {worker_id}: Unexpected error: {e}")
+                    break
+        finally:
+            # Clean up any resources
+            pass
     
     def _collate_batch(self, batch_data):
         """Efficiently collate batch data into tensors with optional pinned memory."""
@@ -111,78 +128,65 @@ class DataLoader:
         # Separate samples and targets
         samples, targets = zip(*batch_data)
         
-        # Convert to numpy arrays first (more efficient than stacking individual tensors)
-        if hasattr(samples[0], 'data'):
-            # If samples are already Tensors, extract data
-            sample_arrays = [s.data for s in samples]
-            target_arrays = [t.data for t in targets]
-        else:
-            # If samples are numpy arrays
-            sample_arrays = list(samples)
-            target_arrays = list(targets)
+        # Convert to arrays
+        sample_arrays = [s.data if hasattr(s, 'data') else s for s in samples]
+        target_arrays = [t.data if hasattr(t, 'data') else t for t in targets]
         
-        # Pin memory optimization: allocate pinned memory if requested
-        if self.pin_memory and IS_CPU:  # Only on CPU with GPU available
-            try:
-                import cupy as cp
-                if cp.cuda.is_available():
-                    # Use pinned memory for efficient GPU transfers
-                    X_data = self._create_pinned_batch(sample_arrays)
-                    y_data = self._create_pinned_batch(target_arrays)
-                else:
-                    # Fallback to regular stacking
-                    X_data = xp.stack(sample_arrays, axis=0)
-                    y_data = xp.stack(target_arrays, axis=0) if len(target_arrays[0].shape) > 0 else xp.array(target_arrays)
-            except ImportError:
-                # CuPy not available, fallback to regular stacking
-                X_data = xp.stack(sample_arrays, axis=0)
-                y_data = xp.stack(target_arrays, axis=0) if len(target_arrays[0].shape) > 0 else xp.array(target_arrays)
-        else:
-            # Regular stacking without pinned memory
-            X_data = xp.stack(sample_arrays, axis=0)
-            y_data = xp.stack(target_arrays, axis=0) if len(target_arrays[0].shape) > 0 else xp.array(target_arrays)
+        # Stack arrays using numpy first
+        X_data = np.stack(sample_arrays, axis=0)
+        y_data = np.stack(target_arrays, axis=0) if target_arrays[0].ndim > 0 else np.array(target_arrays)
         
+        # Use standard pinned memory approach when GPU is available
+        if self.pin_memory and not IS_CPU:
+            X_data = self._pin_memory(X_data)
+            y_data = self._pin_memory(y_data)
+        
+        # Import Tensor locally to avoid serialization issues
+        from neurograd import Tensor
         # Create tensors
         X = Tensor(X_data, dtype=samples[0].dtype if hasattr(samples[0], 'dtype') else X_data.dtype)
         y = Tensor(y_data, dtype=targets[0].dtype if hasattr(targets[0], 'dtype') else y_data.dtype)
             
         return X, y
     
-    def _create_pinned_batch(self, arrays):
-        """Create a batch tensor using pinned memory for faster GPU transfers."""
+    def _pin_memory(self, arr):
+        """Create pinned CPU memory for faster GPU transfers with reusable pool."""
         try:
             import cupy as cp
-            import numpy as np
+            if not cp.cuda.is_available():
+                return arr
+                
+            # Create cache key
+            cache_key = (arr.shape, arr.dtype)
             
-            # Calculate total size needed
-            if len(arrays) == 0:
-                return xp.array([])
+            # Initialize pool if needed
+            if DataLoader._pinned_pool is None:
+                DataLoader._pinned_pool = cp.get_default_pinned_memory_pool()
             
-            # Determine batch shape
-            batch_shape = (len(arrays),) + arrays[0].shape
-            total_bytes = np.prod(batch_shape) * arrays[0].dtype.itemsize
+            # Try to reuse cached pinned array
+            if cache_key in DataLoader._pinned_cache:
+                pinned_array = DataLoader._pinned_cache[cache_key]
+                if pinned_array.nbytes >= arr.nbytes:
+                    # Reshape and copy data
+                    pinned_view = pinned_array[:arr.size].reshape(arr.shape)
+                    pinned_view[...] = arr.flat
+                    return pinned_view.copy()  # Return copy to avoid reference issues
             
-            # Allocate pinned memory
-            pinned_pool = cp.get_default_pinned_memory_pool()
-            pinned_mem = pinned_pool.malloc(total_bytes)
+            # Create new pinned memory
+            pinned_mem = DataLoader._pinned_pool.malloc(arr.nbytes)
+            pinned_array = np.frombuffer(pinned_mem, dtype=arr.dtype).reshape(arr.shape)
+            pinned_array[...] = arr
             
-            # Create numpy array view of pinned memory
-            pinned_array = np.frombuffer(
-                pinned_mem, dtype=arrays[0].dtype, count=np.prod(batch_shape)
-            ).reshape(batch_shape)
-            
-            # Copy data into pinned memory
-            for i, arr in enumerate(arrays):
-                pinned_array[i] = arr
-            
-            # Store reference to prevent garbage collection
-            pinned_array._pinned_mem_ref = pinned_mem
+            # Cache for reuse (limit cache size)
+            if len(DataLoader._pinned_cache) < 10:
+                DataLoader._pinned_cache[cache_key] = pinned_array
+                pinned_array._pinned_mem_ref = pinned_mem  # Prevent GC
             
             return pinned_array
             
-        except Exception:
-            # Fallback to regular stacking if pinned memory allocation fails
-            return xp.stack(arrays, axis=0) if len(arrays[0].shape) > 0 else xp.array(arrays)
+        except ImportError:
+            pass
+        return arr
     def __init__(self, dataset: Dataset, batch_size: int = 32,
                  shuffle: bool = True, seed: Optional[int] = None,
                  num_workers: Optional[int] = None,
@@ -227,6 +231,7 @@ class DataLoader:
         self._index_queues = []
         self._data_queue = None
         self._shutdown_event = None
+        self._worker_load = []  # Track worker load for better balancing
         
         # Performance monitoring
         self._stats = {
@@ -254,14 +259,28 @@ class DataLoader:
         }
 
     def __getitem__(self, idx):
-        """Get a specific batch by index. Enables random.choice(dataloader)."""
+        """Get a specific batch by index with multiprocessing safety."""
         if idx < 0 or idx >= len(self):
             raise IndexError(f"Batch index {idx} out of range [0, {len(self)})")
-        # Get all batch indices for consistent ordering
+            
+        # Always use single-process loading for random access to avoid deadlocks
+        # with worker processes that are designed for sequential iteration
         batches = list(self._batch_indices())
         batch_idxs = batches[idx]
-        # Load the batch synchronously
-        batch_data = [self.dataset[i] for i in batch_idxs]
+        
+        # Load the batch synchronously in main process
+        batch_data = []
+        for i in batch_idxs:
+            try:
+                sample = self.dataset[i]
+                batch_data.append(sample)
+            except Exception as e:
+                print(f"Error loading sample {i} in __getitem__: {e}")
+                continue
+                
+        if not batch_data:
+            raise RuntimeError(f"No valid samples in batch {idx}")
+            
         return self._collate_batch(batch_data)
 
     def _start_workers(self):
@@ -269,9 +288,15 @@ class DataLoader:
         if self.num_workers == 0 or self._workers:
             return
         
-        # Use spawn method for better Windows compatibility
-        ctx = mp.get_context('spawn')
-        self._data_queue = ctx.Queue(maxsize=self.prefetch_batches * 2)
+        # Always use spawn to avoid CUDA context sharing issues
+        # Spawn creates completely isolated processes with clean state
+        try:
+            ctx = mp.get_context('spawn')
+        except RuntimeError:
+            # Fallback to default context if spawn is not available
+            ctx = mp
+        
+        self._data_queue = ctx.Queue(maxsize=max(self.num_workers + self.prefetch_batches, 4))
         self._shutdown_event = ctx.Event()
         
         for worker_id in range(self.num_workers):
@@ -286,36 +311,88 @@ class DataLoader:
             worker.start()
             self._workers.append(worker)
             
-        print(f"[DataLoader] Started {len(self._workers)} worker processes")
+        # Initialize worker load tracking
+        self._worker_load = [0] * self.num_workers
+            
+        print(f"[DataLoader] Started {len(self._workers)} worker processes using spawn context (CUDA-safe)")
 
     def _stop_workers(self):
-        """Stop multiprocessing workers."""
+        """Stop multiprocessing workers with improved safety."""
         if not self._workers:
             return
-            
-        # Send shutdown signals
-        for index_queue in self._index_queues:
+        
+        print(f"[DataLoader] Stopping {len(self._workers)} workers...")
+        
+        # Send shutdown signals to all workers
+        shutdown_success = 0
+        for i, index_queue in enumerate(self._index_queues):
             try:
-                index_queue.put(None, timeout=1)
-            except:
-                pass
+                # Try multiple times to ensure signal is sent
+                for attempt in range(3):
+                    try:
+                        index_queue.put(None, timeout=0.5)
+                        shutdown_success += 1
+                        break
+                    except queue.Full:
+                        if attempt == 2:
+                            print(f"Warning: Could not send shutdown signal to worker {i}")
+                        continue
+            except Exception as e:
+                print(f"Error sending shutdown to worker {i}: {e}")
                 
-        # Wait for workers to finish
-        for worker in self._workers:
-            worker.join(timeout=2)
-            if worker.is_alive():
-                worker.terminate()
-                worker.join()
+        # Wait for workers with increasing timeouts
+        for i, worker in enumerate(self._workers):
+            try:
+                worker.join(timeout=3)  # Increased timeout
+                if worker.is_alive():
+                    print(f"Worker {i} did not terminate gracefully, forcing termination")
+                    worker.terminate()
+                    worker.join(timeout=1)
+                    if worker.is_alive():
+                        print(f"Warning: Worker {i} could not be terminated")
+            except Exception as e:
+                print(f"Error stopping worker {i}: {e}")
                 
-        # Clean up
+        # Clean up resources
         self._workers.clear()
+        self._worker_load.clear()
+        
+        # Close queues safely
+        for queue in self._index_queues:
+            try:
+                # Drain any remaining items
+                while True:
+                    try:
+                        queue.get_nowait()
+                    except queue.Empty:
+                        break
+                queue.close()
+            except Exception:
+                pass
         self._index_queues.clear()
+        
         if self._data_queue:
-            self._data_queue.close()
+            try:
+                # Drain data queue
+                while True:
+                    try:
+                        self._data_queue.get_nowait()
+                    except queue.Empty:
+                        break
+                self._data_queue.close()
+            except Exception:
+                pass
             self._data_queue = None
+            
         self._shutdown_event = None
         
-        print("[DataLoader] Stopped all worker processes")
+        print(f"[DataLoader] Workers stopped (signals sent: {shutdown_success}/{len(self._index_queues)})")
+
+    def _get_best_worker(self):
+        """Get the worker with the lowest current load for better load balancing."""
+        if not self._worker_load:
+            return 0
+        return self._worker_load.index(min(self._worker_load))
 
     def _batch_indices(self):
         n = len(self.dataset)
@@ -336,6 +413,9 @@ class DataLoader:
         if self.num_workers == 0:
             return self._single_process_iter()
         else:
+            # Ensure we don't have any existing workers if not using persistent workers
+            if self._workers and not self.persistent_workers:
+                self._stop_workers()
             return self._multi_process_iter()
             
     def _single_process_iter(self):
@@ -364,25 +444,42 @@ class DataLoader:
             self._start_workers()
             
         try:
-            # Submit initial batches for prefetching
+            # Submit initial batches for prefetching with load balancing
             submitted = 0
             for i in range(min(self.prefetch_batches, len(batches))):
-                self._index_queues[i % self.num_workers].put(batches[i], timeout=self.timeout)
-                submitted += 1
+                worker_id = self._get_best_worker()
+                batch_indices = batches[i]
+                try:
+                    self._index_queues[worker_id].put(batch_indices, timeout=self.timeout)
+                    # Track load by number of samples for accurate balancing
+                    self._worker_load[worker_id] += len(batch_indices)
+                    submitted += 1
+                except queue.Full:
+                    print(f"Warning: Worker {worker_id} queue full, skipping prefetch")
+                    break
                 
             # Process batches
             for batch_idx in range(len(batches)):
                 start_time = time.time()
                 
-                # Submit next batch for prefetching
+                # Submit next batch for prefetching with load balancing
                 if submitted < len(batches):
-                    worker_id = submitted % self.num_workers
-                    self._index_queues[worker_id].put(batches[submitted], timeout=self.timeout)
-                    submitted += 1
+                    worker_id = self._get_best_worker()
+                    try:
+                        self._index_queues[worker_id].put(batches[submitted], timeout=self.timeout)
+                        # Track load by number of samples for accurate balancing
+                        self._worker_load[worker_id] += len(batches[submitted])
+                        submitted += 1
+                    except queue.Full:
+                        print(f"Warning: Worker {worker_id} queue full during iteration")
                 
                 # Get processed batch
                 try:
-                    indices, batch_data = self._data_queue.get(timeout=self.timeout)
+                    worker_id, indices, batch_data = self._data_queue.get(timeout=self.timeout)
+                    # Update worker load accurately - decrement by samples, not batches
+                    if worker_id < len(self._worker_load) and self._worker_load[worker_id] > 0:
+                        self._worker_load[worker_id] -= len(indices)
+                    
                     X, y = self._collate_batch(batch_data)
                     
                     # Update stats
@@ -480,49 +577,41 @@ class ImageFolder(Dataset):
         return self.img_transform(arr)
 
     def _load_image(self, path: str) -> np.ndarray:
-        """Optimized image loading with OpenCV."""
-        mode = (self.img_mode or "RGB").upper()
+        """Optimized image loading with OpenCV and simple format handling."""
+        # Use standard OpenCV loading with simpler conversion logic
+        mode = self.img_mode.upper() if self.img_mode else "RGB"
+        
         if mode in ("L", "GRAY", "GREY", "GRAYSCALE"):
-            flag = cv2.IMREAD_GRAYSCALE
-        elif mode == "RGBA":
-            flag = cv2.IMREAD_UNCHANGED  # preserve alpha if present
+            img = cv2.imread(path, cv2.IMREAD_GRAYSCALE)
+            if img is not None and len(img.shape) == 2:
+                img = img[:, :, np.newaxis]
         else:
-            flag = cv2.IMREAD_COLOR  # BGR
+            img = cv2.imread(path, cv2.IMREAD_COLOR)
+            if img is not None and mode == "RGB":
+                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
+            elif img is not None and mode == "RGBA":
+                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGBA)
         
-        # Avoid EXIF orientation work
-        try:
-            flag |= cv2.IMREAD_IGNORE_ORIENTATION
-        except Exception:
-            pass
-
-        arr = cv2.imread(path, flag)
-        if arr is None:
-            raise ValueError(f"Failed to read image: {path}")
-
-        # Convert channel order to match RGB/RGBA expectations
-        if mode == "RGB" and arr.ndim == 3 and arr.shape[2] == 3:
-            arr = cv2.cvtColor(arr, cv2.COLOR_BGR2RGB)
-        elif mode == "RGBA" and arr.ndim == 3 and arr.shape[2] == 4:
-            arr = cv2.cvtColor(arr, cv2.COLOR_BGRA2RGBA)
-
-        # Resize if requested (cv2 expects (W,H))
+        if img is None:
+            raise ValueError(f"Failed to load image: {path}")
+        
+        # Resize if needed
         if self.img_shape is not None:
-            h, w = self.img_shape
-            arr = cv2.resize(arr, (int(w), int(h)), interpolation=cv2.INTER_LINEAR)
-        
-        if arr.ndim == 2:
-            arr = arr[:, :, None]
+            img = cv2.resize(img, (self.img_shape[1], self.img_shape[0]))
         
+        # Apply transforms
         if self.img_transform:
-            arr = self._apply_img_transform(arr)
+            img = self._apply_img_transform(img)
         
-        if self.chw:
-            arr = np.transpose(arr, (2, 0, 1))  # C,H,W
+        # Convert to CHW if requested
+        if self.chw and img.ndim == 3:
+            img = img.transpose(2, 0, 1)
         
+        # Normalize
         if self.img_normalize:
-            arr = arr.astype(np.float32) / 255.0
+            img = img.astype(np.float32) / 255.0
         
-        return arr
+        return img
 
     def __getitem__(self, idx: int):
         img_path = self.images[idx]
@@ -534,6 +623,8 @@ class ImageFolder(Dataset):
         if self.target_transform:
             target = self.target_transform(target)
 
+        # Import Tensor locally to avoid serialization issues in multiprocessing
+        from neurograd import Tensor
         # Convert to Tensor with dtype specified in init
         return Tensor(image, dtype=self.img_dtype), Tensor(target, dtype=self.target_dtype)
 
@@ -559,4 +650,23 @@ class ImageFolder(Dataset):
 
     def __iter__(self):
         for i in range(len(self)):
-            yield self[i]
\ No newline at end of file
+            yield self[i]
+
+
+
+class OneHotEncoder:
+    def __init__(self, num_classes: int, label_smoothing: float = 0.0):
+        self.num_classes = num_classes
+        self.label_smoothing = label_smoothing
+    
+    def __call__(self, class_idx: int):
+        # Import xp locally to avoid serialization issues in multiprocessing
+        from neurograd import xp
+        one_hot = xp.zeros(self.num_classes, dtype=xp.float32)
+        if self.label_smoothing > 0:
+            smooth_value = self.label_smoothing / self.num_classes
+            one_hot.fill(smooth_value)
+            one_hot[class_idx] = 1.0 - self.label_smoothing + smooth_value
+        else:
+            one_hot[class_idx] = 1.0
+        return one_hot
\ No newline at end of file
